'use strict';

exports.__esModule = true;

var _utils = require('./utils');

var attributeRegex = /\s*(?:[a-zA-Z_][a-zA-Z0-9_]*(?:\s*(?::|-)\s*[a-zA-Z_])?)+\s*/;
var parenthesizedAttributeRegex = (0, _utils.composeRegex)(/\s*(?:\{|\[|\()\s*/, attributeRegex, /\s*(?:\)|\]|\})\s*/);
var attributeAssignmentRegex = /\s*=\s*(?:{|"|')\s*/;

var tagCloses = [{
  regex: /\s*>/
}, {
  regex: /\s*\/\s*>/
}];

var tagAttributes = [].concat(tagCloses, [{
  regex: attributeRegex
}, {
  regex: parenthesizedAttributeRegex
}, {
  regex: attributeRegex,
  alternatives: [].concat(tagCloses, [{
    regex: attributeAssignmentRegex
  }])
}, {
  regex: parenthesizedAttributeRegex,
  alternatives: [].concat(tagCloses, [{
    regex: attributeAssignmentRegex
  }])
}]);

tagAttributes[2].alternatives = tagAttributes;
tagAttributes[3].alternatives = tagAttributes;

var tagMatcher = {
  regex: /</,
  alternatives: [{
    regex: /!--/
  }, {
    regex: /\s*(?:[a-zA-Z_][a-zA-Z0-9_]*(?:\s*(?::|->?|\.)\s*[a-zA-Z_])?)+\s*/,
    alternatives: tagAttributes
  }]
};

var isTagOpen = function isTagOpen(str) {
  return (0, _utils.stringMatch)(str, tagMatcher);
};

var normalizeNewLines = function normalizeNewLines(str) {
  return str.replace(/\s*\n\s*/g, ' ');
};

var escape = function escape(char, str) {
  return str.replace(new RegExp('(' + char + ')', 'g'), '\\$1');
};

var escapeQuotes = escape.bind(null, '"');

var getTagName = function getTagName(sel) {
  if (sel.indexOf('u8"') === 0) {
    return sel.substring(0, sel.length - 1).replace('u8"', '');
  }
  return sel;
};

var filterByType = function filterByType(type, vnode) {
  return vnode.data.filter(function (attribute) {
    return attribute.type === type;
  }).filter(function (obj, index, arr) {
    return arr.map(function (mapObj) {
      return mapObj.name;
    }).lastIndexOf(obj.name) === index;
  });
};

var filterAttrs = filterByType.bind(null, 'attr');
var filterProps = filterByType.bind(null, 'prop');
var filterCallbacks = filterByType.bind(null, 'callback');

var createMaps = function createMaps(maps) {
  return maps.map(function (map) {
    var res = '';
    if (map.values.length !== 0) {
      res += map.id + ' {';
      res += map.values.map(function (attr) {
        return '{' + attr.name + ', ' + attr.value + '}';
      }).join(', ');
      res += '}';
    }
    return res;
  }).filter(function (map) {
    return map !== '';
  }).join(', ');
};

var aggregateStrings = function aggregateStrings(vnodes) {
  if (vnodes.length === 1) return vnodes[0];
  return {
    type: 'string',
    aggregated: true,
    value: vnodes.map(function (vnode, index) {
      var value = vnode.value;


      if (vnode.type === 'CPXText') {
        if (index === 0) {
          value = value.replace(/^\s+/, '');
        } else if (index === vnodes.length - 1) {
          value = value.replace(/\s+$/, '');
        }
        return 'u8"' + value + '"';
      } else if (vnode.type === 'string') {
        return 'std::string(' + value.trim() + ')';
      }
      return '';
    }).join(' + ')
  };
};

var aggregateNodes = function aggregateNodes(vnodes) {
  var computedVnodes = [];
  var strings = [];
  for (var i = 0; i < vnodes.length; i += 1) {
    var child = vnodes[i];
    if (child.type === 'CPXText' || child.type === 'string') {
      strings.push(child);
    } else {
      if (strings.length !== 0) {
        computedVnodes.push(aggregateStrings(strings));
        strings = [];
      }
      computedVnodes.push(child);
    }
  }
  if (strings.length !== 0) {
    computedVnodes.push(aggregateStrings(strings));
  }
  return computedVnodes;
};

var createVNode = function createVNode(data) {
  var vnode = void 0;
  if (data.type === 'VNode') return data.value;
  if (data.type === 'string') return 'asmdom::h(' + data.value.trim() + ', true)';
  if (data.type === 'CPXText') return 'asmdom::h(u8"' + data.value.trim() + '", true)';
  if (data.type === 'CPXComment') return 'asmdom::h(u8"!", std::string(u8"' + escapeQuotes(data.value) + '"))';
  if (data.type === 'CPXElement') {
    vnode = 'asmdom::h(' + data.sel;

    if (data.data !== undefined && data.data.length !== 0) {
      vnode += ', asmdom::Data (';
      vnode += createMaps([{ id: 'asmdom::Attrs', values: filterAttrs(data) }, { id: 'asmdom::Props', values: filterProps(data) }, { id: 'asmdom::Callbacks', values: filterCallbacks(data) }]);
      vnode += ')';
    }

    if (data.children !== undefined) {
      var children = aggregateNodes(data.children.filter(function (child) {
        return child.type !== 'comment';
      }));

      if (children.length === 1) {
        vnode += ', ';
        if (children[0].type === 'CPXText') {
          vnode += 'std::string(u8"' + children[0].value.trim() + '")';
        } else if (children[0].type === 'string') {
          vnode += children[0].aggregated !== true ? 'std::string(' + children[0].value.trim() + ')' : children[0].value.trim();
        } else {
          vnode += createVNode(children[0]);
        }
      } else if (children.length > 1) {
        vnode += ', asmdom::Children {' + children.map(createVNode).join(', ') + '}';
      }
    }

    vnode += ')';
  }
  return vnode;
};

exports['default'] = {
  createVNode: createVNode,
  normalizeNewLines: normalizeNewLines,
  escapeQuotes: escapeQuotes,
  getTagName: getTagName,
  isTagOpen: isTagOpen
};